"""Unit tests for event handler module."""

from datetime import timedelta
from typing import Generator, Sequence
from unittest.mock import Mock, patch

import pytest

from async_upnp_client.client import UpnpService, UpnpStateVariable
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import HttpRequest
from async_upnp_client.event_handler import UpnpEventHandlerRegister

from .conftest import RESPONSE_MAP, UpnpTestNotifyServer, UpnpTestRequester


@pytest.fixture
def patched_local_ip() -> Generator:
    """Patch get_local_ip to `'192.168.1.2"`."""
    with patch("async_upnp_client.event_handler.get_local_ip") as mock:
        yield mock


@pytest.mark.asyncio
async def test_subscribe() -> None:
    """Test subscribing to a UpnpService."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler

    service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
    sid, timeout = await event_handler.async_subscribe(service)
    assert event_handler.service_for_sid("uuid:dummy") == service
    assert sid == "uuid:dummy"
    assert timeout == timedelta(seconds=300)
    assert event_handler.callback_url == "http://192.168.1.2:8090/notify"


@pytest.mark.asyncio
async def test_subscribe_renew() -> None:
    """Test renewing an existing subscription to a UpnpService."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler

    service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
    sid, timeout = await event_handler.async_subscribe(service)
    assert sid == "uuid:dummy"
    assert event_handler.service_for_sid("uuid:dummy") == service
    assert timeout == timedelta(seconds=300)

    sid, timeout = await event_handler.async_resubscribe(service)
    assert event_handler.service_for_sid("uuid:dummy") == service
    assert sid == "uuid:dummy"
    assert timeout == timedelta(seconds=300)


@pytest.mark.asyncio
async def test_unsubscribe() -> None:
    """Test unsubscribing from a UpnpService."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler

    service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
    sid, timeout = await event_handler.async_subscribe(service)
    assert event_handler.service_for_sid("uuid:dummy") == service
    assert sid == "uuid:dummy"
    assert timeout == timedelta(seconds=300)

    old_sid = await event_handler.async_unsubscribe(service)
    assert event_handler.service_for_sid("uuid:dummy") is None
    assert old_sid == "uuid:dummy"


@pytest.mark.asyncio
async def test_on_notify_upnp_event() -> None:
    """Test handling of a UPnP event."""
    changed_vars: Sequence[UpnpStateVariable] = []

    def on_event(_self: UpnpService, changed_state_variables: Sequence[UpnpStateVariable]) -> None:
        nonlocal changed_vars
        changed_vars = changed_state_variables

    requester = UpnpTestRequester(RESPONSE_MAP)
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
    service.on_event = on_event
    await event_handler.async_subscribe(service)

    headers = {
        "NT": "upnp:event",
        "NTS": "upnp:propchange",
        "SID": "uuid:dummy",
    }
    body = """
<e:propertyset xmlns:e="urn:schemas-upnp-org:event-1-0">
    <e:property>
        <Volume>60</Volume>
    </e:property>
</e:propertyset>
"""

    http_request = HttpRequest("NOTIFY", "http://dlna_dmr:1234/upnp/event/RenderingControl1", headers, body)
    result = await event_handler.handle_notify(http_request)
    assert result == 200

    assert len(changed_vars) == 1

    state_var = service.state_variable("Volume")
    assert state_var.value == 60


@pytest.mark.asyncio
async def test_register_device(patched_local_ip: Mock) -> None:
    """Test registering a device with a UpnpEventHandlerRegister."""
    # pylint: disable=redefined-outer-name
    requester = UpnpTestRequester(RESPONSE_MAP)
    register = UpnpEventHandlerRegister(requester, UpnpTestNotifyServer)
    patched_local_ip.return_value = "192.168.1.2"

    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")

    event_handler = await register.async_add_device(device)
    assert event_handler is not None
    assert event_handler.callback_url == "http://192.168.1.2:0/notify"
    assert register.has_event_handler_for_device(device)


@pytest.mark.asyncio
async def test_register_device_different_source_address(patched_local_ip: Mock) -> None:
    """Test registering two devices with different source IPs with a UpnpEventHandlerRegister."""
    # pylint: disable=redefined-outer-name
    requester = UpnpTestRequester(RESPONSE_MAP)
    register = UpnpEventHandlerRegister(requester, UpnpTestNotifyServer)
    factory = UpnpFactory(requester)

    patched_local_ip.return_value = "192.168.1.2"
    device_1 = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    event_handler_1 = await register.async_add_device(device_1)
    assert event_handler_1 is not None
    assert event_handler_1.callback_url == "http://192.168.1.2:0/notify"

    patched_local_ip.return_value = "192.168.2.2"
    device_2 = await factory.async_create_device("http://igd:1234/device.xml")
    event_handler_2 = await register.async_add_device(device_2)
    assert event_handler_2 is not None
    assert event_handler_2.callback_url == "http://192.168.2.2:0/notify"


@pytest.mark.asyncio
async def test_remove_device(patched_local_ip: Mock) -> None:
    """Test removing a device from a UpnpEventHandlerRegister."""
    # pylint: disable=redefined-outer-name
    requester = UpnpTestRequester(RESPONSE_MAP)
    register = UpnpEventHandlerRegister(requester, UpnpTestNotifyServer)
    factory = UpnpFactory(requester)

    patched_local_ip.return_value = "192.168.1.2"
    device_1 = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    device_2 = await factory.async_create_device("http://igd:1234/device.xml")

    event_handler_1 = await register.async_add_device(device_1)
    event_handler_2 = await register.async_add_device(device_2)
    assert event_handler_1 is event_handler_2

    removed_event_handler_1 = await register.async_remove_device(device_1)
    assert removed_event_handler_1 is None  # UpnpEventHandler is still being used
    removed_event_handler_2 = await register.async_remove_device(device_2)
    assert removed_event_handler_2 is event_handler_1



"""Unit tests for the DLNA DMR profile."""

import asyncio
import time
from typing import Sequence
from unittest import mock

import defusedxml.ElementTree
import pytest
from didl_lite import didl_lite

from async_upnp_client.client import UpnpService, UpnpStateVariable
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import HttpRequest, HttpResponse
from async_upnp_client.profiles.dlna import (
    DmrDevice,
    _parse_last_change_event,
    dlna_handle_notify_last_change,
    split_commas,
)

from ..conftest import RESPONSE_MAP, UpnpTestNotifyServer, UpnpTestRequester, read_file

AVT_NOTIFY_HEADERS = {
    "NT": "upnp:event",
    "NTS": "upnp:propchange",
    "SID": "uuid:dummy-avt1",
}

AVT_CURRENT_TRANSPORT_ACTIONS_NOTIFY_BODY_FMT = """
<e:propertyset xmlns:e="urn:schemas-upnp-org:event-1-0">
    <e:property>
        <LastChange>
            &lt;Event xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/AVT/&quot;&gt;
                &lt;InstanceID val=&quot;0&quot;&gt;
                    &lt;CurrentTransportActions val=&quot;{actions}&quot;/&gt;
                    &lt;/InstanceID&gt;
            &lt;/Event&gt;
        </LastChange>
    </e:property>
</e:propertyset>
"""


def assert_xml_equal(left: defusedxml.ElementTree, right: defusedxml.ElementTree) -> None:
    """Check two XML trees are equal."""
    assert left.tag == right.tag
    assert left.text == right.text
    assert left.tail == right.tail
    assert left.attrib == right.attrib
    assert len(left) == len(right)
    for left_child, right_child in zip(left, right):
        assert_xml_equal(left_child, right_child)


def test_parse_last_change_event() -> None:
    """Test parsing a last change event."""
    data = """<Event xmlns="urn:schemas-upnp-org:metadata-1-0/AVT/">
<InstanceID val="0"><TransportState val="PAUSED_PLAYBACK"/></InstanceID>
</Event>"""
    assert _parse_last_change_event(data) == {"0": {"TransportState": "PAUSED_PLAYBACK"}}


def test_parse_last_change_event_multiple_instances() -> None:
    """Test parsing a last change event with multiple instance."""
    data = """<Event xmlns="urn:schemas-upnp-org:metadata-1-0/AVT/">
<InstanceID val="0"><TransportState val="PAUSED_PLAYBACK"/></InstanceID>
<InstanceID val="1"><TransportState val="PLAYING"/></InstanceID>
</Event>"""
    assert _parse_last_change_event(data) == {
        "0": {"TransportState": "PAUSED_PLAYBACK"},
        "1": {"TransportState": "PLAYING"},
    }


def test_parse_last_change_event_multiple_channels() -> None:
    """Test parsing a last change event with multiple channels."""
    data = """<Event xmlns="urn:schemas-upnp-org:metadata-1-0/AVT/">
<InstanceID val="0">
  <Volume channel="Master" val="10"/>
  <Volume channel="Left" val="20"/>
  <Volume channel="Right" val="30"/>
</InstanceID>
</Event>"""
    assert _parse_last_change_event(data) == {
        "0": {"Volume": "10"},
    }


def test_parse_last_change_event_invalid_xml() -> None:
    """Test parsing an invalid (non valid XML) last change event."""
    data = """<Event xmlns="urn:schemas-upnp-org:metadata-1-0/AVT/">
<InstanceID val="0"><TransportState val="PAUSED_PLAYBACK"></InstanceID>
</Event>"""
    assert _parse_last_change_event(data) == {"0": {"TransportState": "PAUSED_PLAYBACK"}}


@pytest.mark.parametrize(
    "value, expected",
    (
        ("", []),
        (",", []),
        (", ,", []),
        (
            "http-get:*:audio/mp3:*,http-get:*:audio/mp4:*,http-get:*:audio/x-m4a:*",
            [
                "http-get:*:audio/mp3:*",
                "http-get:*:audio/mp4:*",
                "http-get:*:audio/x-m4a:*",
            ],
        ),
        (
            "http-get:*:audio/mp3:*,http-get:*:audio/mp4:*,http-get:*:audio/x-m4a:*,",
            [
                "http-get:*:audio/mp3:*",
                "http-get:*:audio/mp4:*",
                "http-get:*:audio/x-m4a:*",
            ],
        ),
    ),
)
def test_split_commas(value: str, expected: list[str]) -> None:
    """Test splitting comma separated value lists."""
    actual = split_commas(value)
    assert actual == expected


@pytest.mark.asyncio
async def test_on_notify_dlna_event() -> None:
    """Test handling an event.."""
    changed_vars: list[UpnpStateVariable] = []

    def on_event(_self: UpnpService, changed_state_variables: Sequence[UpnpStateVariable]) -> None:
        nonlocal changed_vars
        changed_vars += changed_state_variables

        assert changed_state_variables
        if changed_state_variables[0].name == "LastChange":
            last_change = changed_state_variables[0]
            assert last_change.name == "LastChange"

            dlna_handle_notify_last_change(last_change)

    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
    service.on_event = on_event
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    await event_handler.async_subscribe(service)

    headers = {
        "NT": "upnp:event",
        "NTS": "upnp:propchange",
        "SID": "uuid:dummy",
    }
    body = """
<e:propertyset xmlns:e="urn:schemas-upnp-org:event-1-0">
    <e:property>
        <LastChange>
            &lt;Event xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/RCS/&quot;&gt;
                &lt;InstanceID val=&quot;0&quot;&gt;
                    &lt;Mute channel=&quot;Master&quot; val=&quot;0&quot;/&gt;
                    &lt;Volume channel=&quot;Master&quot; val=&quot;50&quot;/&gt;
                    &lt;/InstanceID&gt;
            &lt;/Event&gt;
        </LastChange>
    </e:property>
</e:propertyset>
"""

    http_request = HttpRequest("NOTIFY", "http://dlna_dmr:1234/upnp/event/RenderingControl1", headers, body)
    result = await event_handler.handle_notify(http_request)
    assert result == 200

    assert len(changed_vars) == 3

    state_var = service.state_variable("Volume")
    assert state_var.value == 50


@pytest.mark.asyncio
async def test_wait_for_can_play_evented() -> None:
    """Test async_wait_for_can_play with a variable change event."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = DmrDevice(device, event_handler=event_handler)
    await profile.async_subscribe_services()

    # Send a NOTIFY of CurrentTransportActions without Play
    http_request = HttpRequest(
        "NOTIFY",
        "http://dlna_dmr:1234/upnp/event/AVTransport1",
        AVT_NOTIFY_HEADERS,
        AVT_CURRENT_TRANSPORT_ACTIONS_NOTIFY_BODY_FMT.format(actions="Stop"),
    )
    result = await event_handler.handle_notify(http_request)
    assert result == 200

    # Should not be able to play yet
    assert not profile.can_play

    # Trigger variable change event in 0.1 seconds, less than the sleep time of
    # the wait loop
    async def delayed_notify() -> None:
        await asyncio.sleep(0.1)
        # Send NOTIFY of change to CurrentTransportActions
        http_request = HttpRequest(
            "NOTIFY",
            "http://dlna_dmr:1234/upnp/event/AVTransport1",
            AVT_NOTIFY_HEADERS,
            AVT_CURRENT_TRANSPORT_ACTIONS_NOTIFY_BODY_FMT.format(actions="Pause,Play"),
        )
        result = await event_handler.handle_notify(http_request)
        assert result == 200

    loop = asyncio.get_event_loop()
    notify_task = loop.create_task(delayed_notify())
    assert notify_task

    # Call async_wait_for_can_play and check it returned shortly after notification
    started = time.monotonic()
    await profile.async_wait_for_can_play()
    waited_time = time.monotonic() - started

    assert 0.1 <= waited_time <= 0.5

    assert profile.can_play

    await profile.async_unsubscribe_services()


@pytest.mark.asyncio
async def test_wait_for_can_play_polled() -> None:
    """Test async_wait_for_can_play polling state variables."""
    requester = UpnpTestRequester(RESPONSE_MAP)

    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    profile = DmrDevice(device, event_handler=None)

    # Polling of CurrentTransportActions does not contain "Play" yet
    requester.response_map[("POST", "http://dlna_dmr:1234/upnp/control/AVTransport1")] = HttpResponse(
        200,
        {},
        read_file("dlna/dmr/action_GetCurrentTransportActions_Stop.xml"),
    )
    # Force update of CurrentTransportActions
    # pylint: disable=protected-access
    await profile._async_poll_state_variables("AVT", ["GetCurrentTransportActions"], InstanceID=0)

    # Should not be able to play yet
    assert not profile.can_play

    # Polling of CurrentTransportActions now contains "Play"
    requester.response_map[("POST", "http://dlna_dmr:1234/upnp/control/AVTransport1")] = HttpResponse(
        200,
        {},
        read_file("dlna/dmr/action_GetCurrentTransportActions_PlaySeek.xml"),
    )

    # Call async_wait_for_can_play and check it returned after polling
    started = time.monotonic()
    await profile.async_wait_for_can_play()
    waited_time = time.monotonic() - started

    assert 0.1 <= waited_time <= 1.0

    assert profile.can_play


@pytest.mark.asyncio
async def test_wait_for_can_play_timeout() -> None:
    """Test async_wait_for_can_play times out waiting for ability to play."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    profile = DmrDevice(device, event_handler=None)

    # Polling of CurrentTransportActions does not contain "Play" yet
    requester.response_map[("POST", "http://dlna_dmr:1234/upnp/control/AVTransport1")] = HttpResponse(
        200,
        {},
        read_file("dlna/dmr/action_GetCurrentTransportActions_Stop.xml"),
    )
    # Force update of CurrentTransportActions
    # pylint: disable=protected-access
    await profile._async_poll_state_variables("AVT", ["GetCurrentTransportActions"], InstanceID=0)

    # Should not be able to play
    assert not profile.can_play

    # Call async_wait_for_can_play with a shorter timeout (to not delay tests too long)
    started = time.monotonic()
    await profile.async_wait_for_can_play(max_wait_time=0.5)
    waited_time = time.monotonic() - started

    assert 0.5 <= waited_time <= 1.5

    assert not profile.can_play


@pytest.mark.asyncio
async def test_fetch_headers() -> None:
    """Test _fetch_headers when the server supports HEAD, GET with range, or just GET."""
    # pylint: disable=protected-access
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    profile = DmrDevice(device, event_handler=None)

    media_url = "http://dlna_dms:4321/object/file_1222"
    fetch_headers = {"GetContentFeatures.dlna.org": "1"}
    expected_response_headers = {"Content-Length": "1024", "Content-Type": "audio/mpeg"}

    # When HEAD works
    with mock.patch.object(profile.profile_device.requester, "async_http_request") as ahr_mock:
        ahr_mock.side_effect = [HttpResponse(200, expected_response_headers, "")]
        headers = await profile._fetch_headers(media_url, fetch_headers)
        ahr_mock.assert_awaited_once_with(HttpRequest("HEAD", media_url, fetch_headers, None))
        assert headers == expected_response_headers

    # HEAD method is not allowed, but GET with Range works
    with mock.patch.object(profile.profile_device.requester, "async_http_request") as ahr_mock:
        ranged_response_headers = dict(expected_response_headers)
        ranged_response_headers["Content-Range"] = "bytes 0-0/1024"
        ahr_mock.side_effect = [
            HttpResponse(405, expected_response_headers, ""),
            HttpResponse(200, ranged_response_headers, ""),
        ]
        headers = await profile._fetch_headers(media_url, fetch_headers)
        assert ahr_mock.await_args_list == [
            mock.call(HttpRequest("HEAD", media_url, fetch_headers, None)),
            mock.call(HttpRequest("GET", media_url, dict(fetch_headers, Range="bytes=0-0"), None)),
        ]
        assert headers == ranged_response_headers

    # HEAD method and GET with Range is not allowed, but plain GET works
    with mock.patch.object(profile.profile_device.requester, "async_http_request") as ahr_mock:
        # Different headers for working response, to check correct thing returned
        get_headers = dict(expected_response_headers)
        get_headers["Content-Length"] = "2"
        ahr_mock.side_effect = [
            HttpResponse(405, expected_response_headers, ""),
            HttpResponse(405, expected_response_headers, ""),
            HttpResponse(200, get_headers, ""),
        ]
        headers = await profile._fetch_headers(media_url, fetch_headers)
        assert ahr_mock.await_args_list == [
            mock.call(HttpRequest("HEAD", media_url, fetch_headers, None)),
            mock.call(HttpRequest("GET", media_url, dict(fetch_headers, Range="bytes=0-0"), None)),
            mock.call(HttpRequest("GET", media_url, fetch_headers, None)),
        ]
        assert headers == get_headers

    # HTTP 404 should bail early
    with mock.patch.object(profile.profile_device.requester, "async_http_request") as ahr_mock:
        ahr_mock.side_effect = [
            HttpResponse(404, expected_response_headers, ""),
            HttpResponse(405, expected_response_headers, ""),
            HttpResponse(200, expected_response_headers, ""),
        ]
        headers = await profile._fetch_headers(media_url, fetch_headers)
        ahr_mock.assert_called_once_with(HttpRequest("HEAD", media_url, fetch_headers, None))
        assert headers is None

    # Repeated server failures should give no headers
    with mock.patch.object(profile.profile_device.requester, "async_http_request") as ahr_mock:
        # Different headers for working response, to check correct thing returned
        ahr_mock.return_value = HttpResponse(500, {}, "")
        headers = await profile._fetch_headers(media_url, fetch_headers)
        assert ahr_mock.await_args_list == [
            mock.call(HttpRequest("HEAD", media_url, fetch_headers, None)),
            mock.call(HttpRequest("GET", media_url, dict(fetch_headers, Range="bytes=0-0"), None)),
            mock.call(HttpRequest("GET", media_url, fetch_headers, None)),
        ]
        assert headers is None


@pytest.mark.asyncio
async def test_construct_play_media_metadata_types() -> None:
    """Test various MIME and UPnP type options for construct_play_media_metadata."""
    # pylint: disable=too-many-statements
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = DmrDevice(device, event_handler=event_handler)

    media_url = "http://dlna_dms:4321/object/file_1222"
    media_title = "Test music"

    # No server to supply DLNA headers
    metadata_xml = await profile.construct_play_media_metadata(media_url, media_title)
    # Sanity check that didl_lite is giving expected XML
    expected_xml = defusedxml.ElementTree.fromstring(
        """<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
 xmlns:dc="http://purl.org/dc/elements/1.1/"
 xmlns:sec="http://www.sec.co.kr/"
 xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">
<item id="0" parentID="-1" restricted="false">
<dc:title>Test music</dc:title>
<upnp:class>object.item</upnp:class>
<res protocolInfo="http-get:*:application/octet-stream:*">
http://dlna_dms:4321/object/file_1222
</res>
</item>
</DIDL-Lite>""".replace("\n", "")
    )
    assert_xml_equal(defusedxml.ElementTree.fromstring(metadata_xml), expected_xml)

    metadata = didl_lite.from_xml_string(metadata_xml)[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item"
    assert metadata.res
    assert metadata.res is metadata.resources
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:application/octet-stream:*"

    metadata = didl_lite.from_xml_string(await profile.construct_play_media_metadata(media_url + ".mp3", media_title))[
        0
    ]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.audioItem"
    assert metadata.res[0].uri == media_url + ".mp3"
    assert metadata.res[0].protocol_info == "http-get:*:audio/mpeg:*"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(media_url, media_title, default_mime_type="video/test-mime")
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/test-mime:*"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(media_url, media_title, default_upnp_class="object.item.imageItem")
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.imageItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:application/octet-stream:*"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(media_url, media_title, override_mime_type="video/test-mime")
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/test-mime:*"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url, media_title, override_upnp_class="object.item.imageItem"
        )
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.imageItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:application/octet-stream:*"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url,
            media_title,
            override_dlna_features="DLNA_OVERRIDE_FEATURES",
        )
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:application/octet-stream:DLNA_OVERRIDE_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url,
            media_title,
            override_mime_type="video/test-mime",
            override_dlna_features="DLNA_OVERRIDE_FEATURES",
        )
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/test-mime:DLNA_OVERRIDE_FEATURES"

    # Media server supplies media information for HEAD requests
    requester.response_map[("HEAD", media_url)] = HttpResponse(
        200,
        {
            "ContentFeatures.dlna.org": "DLNA_SERVER_FEATURES",
            "Content-Type": "video/server-mime",
        },
        "",
    )
    requester.response_map[("HEAD", media_url + ".mp3")] = HttpResponse(
        200,
        {
            "ContentFeatures.dlna.org": "DLNA_SERVER_FEATURES",
            "Content-Type": "video/server-mime",
        },
        "",
    )

    metadata = didl_lite.from_xml_string(await profile.construct_play_media_metadata(media_url, media_title))[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/server-mime:DLNA_SERVER_FEATURES"

    metadata = didl_lite.from_xml_string(await profile.construct_play_media_metadata(media_url + ".mp3", media_title))[
        0
    ]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url + ".mp3"
    assert metadata.res[0].protocol_info == "http-get:*:video/server-mime:DLNA_SERVER_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(media_url, media_title, default_mime_type="video/test-mime")
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/server-mime:DLNA_SERVER_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(media_url, media_title, default_upnp_class="object.item.imageItem")
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/server-mime:DLNA_SERVER_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(media_url, media_title, override_mime_type="image/test-mime")
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.imageItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:image/test-mime:DLNA_SERVER_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url, media_title, override_upnp_class="object.item.imageItem"
        )
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.imageItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/server-mime:DLNA_SERVER_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url,
            media_title,
            override_dlna_features="DLNA_OVERRIDE_FEATURES",
        )
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.videoItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:video/server-mime:DLNA_OVERRIDE_FEATURES"

    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url,
            media_title,
            override_mime_type="image/test-mime",
            override_dlna_features="DLNA_OVERRIDE_FEATURES",
        )
    )[0]
    assert metadata.title == media_title
    assert metadata.upnp_class == "object.item.imageItem"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:image/test-mime:DLNA_OVERRIDE_FEATURES"


@pytest.mark.asyncio
async def test_construct_play_media_metadata_meta_data() -> None:
    """Test meta_data values for construct_play_media_metadata."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = DmrDevice(device, event_handler=event_handler)

    media_url = "http://dlna_dms:4321/object/file_1222.mp3"
    media_title = "Test music"
    meta_data = {
        "title": "Test override title",  # Should override media_title parameter
        "description": "Short test description",  # In base audioItem class
        "artist": "Test singer",
        "album": "Test album",
        "originalTrackNumber": 3,  # Should be converted to lower_camel_case
    }

    # No server information about media type or contents

    # Without specifying UPnP class, only generic types lacking certain values are used
    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url,
            media_title,
            meta_data=meta_data,
        )
    )[0]
    assert metadata.upnp_class == "object.item.audioItem"
    assert metadata.title == "Test override title"
    assert metadata.description == "Short test description"
    assert not hasattr(metadata, "artist")
    assert not hasattr(metadata, "album")
    assert not hasattr(metadata, "original_track_number")
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:audio/mpeg:*"

    # Set the UPnP class correctly
    metadata = didl_lite.from_xml_string(
        await profile.construct_play_media_metadata(
            media_url,
            media_title,
            override_upnp_class="object.item.audioItem.musicTrack",
            meta_data=meta_data,
        )
    )[0]
    assert metadata.upnp_class == "object.item.audioItem.musicTrack"
    assert metadata.title == "Test override title"
    assert metadata.description == "Short test description"
    assert metadata.artist == "Test singer"
    assert metadata.album == "Test album"
    assert metadata.original_track_number == "3"
    assert metadata.res[0].uri == media_url
    assert metadata.res[0].protocol_info == "http-get:*:audio/mpeg:*"



#!/usr/bin/env python3
"""Dummy mediaseerver."""
# Instructions:
# - Change `SOURCE``. When using IPv6, be sure to set the scope_id, the last value in the tuple.
# - Run this module.
# - Run upnp-client (change IP to your own IP):
#    upnp-client call-action 'http://0.0.0.0:8000/device.xml' \
#                WANCIC/GetTotalPacketsReceived

import asyncio
import logging
import xml.etree.ElementTree as ET
from time import time
from datetime import datetime

from async_upnp_client.client import UpnpRequester, UpnpStateVariable
from async_upnp_client.const import (
    DeviceInfo,
    ServiceInfo,
)

from async_upnp_client.server import (
    UpnpServer,
    UpnpServerDevice,
    UpnpServerService,
    callable_action,
    create_state_var,
    create_event_var)

logging.basicConfig(level=logging.DEBUG)
LOGGER = logging.getLogger("dummy_mediaserver")
LOGGER_SSDP_TRAFFIC = logging.getLogger("async_upnp_client.traffic")
LOGGER_SSDP_TRAFFIC.setLevel(logging.WARNING)
SOURCE = ("192.168.1.85", 0)  # Your IP here!
# SOURCE = ("fe80::215:5dff:fe3e:6d23", 0, 0, 6)  # Your IP here!
HTTP_PORT = 8000


class ContentDirectoryService(UpnpServerService):
    """DLNA Content Directory."""

    SERVICE_DEFINITION = ServiceInfo(
        service_id="urn:upnp-org:serviceId:ContentDirectory",
        service_type="urn:schemas-upnp-org:service:ContentDirectory:2",
        control_url="/upnp/control/ContentDirectory",
        event_sub_url="/upnp/event/ContentDirectory",
        scpd_url="/ContentDirectory.xml",
        xml=ET.Element("server_service"),
    )

    STATE_VARIABLE_DEFINITIONS = {
        "SearchCapabilities": create_state_var("string"),
        "SortCapabilities": create_state_var("string"),
        "SystemUpdateID": create_event_var("ui4", max_rate=0.2),
        "FeatureList": create_state_var("string",
            default="""<?xml version="1.0" encoding="UTF-8"?>
<Features
 xmlns="urn:schemas-upnp-org:av:avs"
 xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 xsi:schemaLocation="
 urn:schemas-upnp-org:av:avs
 http://www.upnp.org/schemas/av/avs-v1-20060531.xsd">
</Features>"""),
        "A_ARG_TYPE_BrowseFlag": create_state_var("string",
            allowed=["BrowseMetadata", "BrowseDirectChildren"]),
        "A_ARG_TYPE_Filter": create_state_var("string"),
        "A_ARG_TYPE_ObjectID": create_state_var("string"),
        "A_ARG_TYPE_Count": create_state_var("ui4"),
        "A_ARG_TYPE_Index": create_state_var("ui4"),
        "A_ARG_TYPE_SortCriteria": create_state_var("string"),
        ###
        "A_ARG_TYPE_Result": create_state_var("string"),
        "A_ARG_TYPE_UpdateID": create_state_var("ui4"),
        "A_ARG_TYPE_Count_NumberReturned": create_state_var("ui4"),
        "A_ARG_TYPE_Count_TotalMatches": create_state_var("ui4"),
    }

    @callable_action(
        name="Browse",
        in_args={
            "BrowseFlag": "A_ARG_TYPE_BrowseFlag",
            "Filter": "A_ARG_TYPE_Filter",
            "ObjectID": "A_ARG_TYPE_ObjectID",
            "RequestedCount": "A_ARG_TYPE_Count",
            "SortCriteria": "A_ARG_TYPE_SortCriteria",
            "StartingIndex": "A_ARG_TYPE_Index",
            },
        out_args={
            "Result": "A_ARG_TYPE_Result",
            "NumberReturned": "A_ARG_TYPE_Count_NumberReturned",
            "TotalMatches": "A_ARG_TYPE_Count_TotalMatches",
            "UpdateID": "A_ARG_TYPE_UpdateID",
        },
    )
    async def browse(self, BrowseFlag: str, Filter: str, ObjectID: str, StartingIndex: int,
                     RequestedCount: int, SortCriteria: str) -> dict[str, UpnpStateVariable]:
        """Browse media."""
        root =  ET.Element("DIDL-Lite", {
             'xmlns:dc': 'http://purl.org/dc/elements/1.1/',
             'xmlns:upnp': 'urn:schemas-upnp-org:metadata-1-0/upnp/',
             'DIDL-Lite': 'urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/'})
        child = ET.Element('item', {'id': f'100', 'restricted': '0'})
        ET.SubElement(child, 'dc:title').text = "Item 1"
        ET.SubElement(child, 'dc:date').text = datetime.now().isoformat()

        root.append(child)
        xml = ET.tostring(root).decode()
        return {
            "Result": xml,
            "NumberReturned": 1,
            "TotalMatches": 2,
            }

    @callable_action(
        name="GetSearchCapabilities",
        in_args={},
        out_args={
            "SearchCaps": "SearchCapabilities",
        },
    )
    async def GetSearchCapabilities(self) -> dict[str, UpnpStateVariable]:
        """Browse media."""
        return {
            "SearchCaps": self.state_variable("SearchCapabilities"),
        }

    @callable_action(
        name="GetSortCapabilities",
        in_args={},
        out_args={
            "SortCaps": "SortCapabilities",
        },
    )
    async def GetSortCapabilities(self) -> dict[str, UpnpStateVariable]:
        """Browse media."""
        return {
            "SortCaps": self.state_variable("SortCapabilities"),
        }
    @callable_action(
        name="GetFeatureList",
        in_args={},
        out_args={
            "FeatureList": "FeatureList",
        },
    )
    async def GetFeatureList(self) -> dict[str, UpnpStateVariable]:
        """Browse media."""
        return {
            "FeatureList": self.state_variable("FeatureList"),
        }
    @callable_action(
        name="GetSystemUpdateID",
        in_args={},
        out_args={
            "Id": "SystemUpdateID",
        },
    )
    async def GetSystemUpdateID(self) -> dict[str, UpnpStateVariable]:
        """Browse media."""
        return {
            "Id": self.state_variable("SystemUpdateID"),
        }


class MediaServerDevice(UpnpServerDevice):
    """Media Server Device."""

    DEVICE_DEFINITION = DeviceInfo(
        device_type=":urn:schemas-upnp-org:device:MediaServer:2",
        friendly_name="Media Server v1",
        manufacturer="Steven",
        manufacturer_url=None,
        model_name="MediaServer v1",
        model_url=None,
        udn="uuid:1cd38bfe-3c10-403e-a97f-2bc5c1652b9a",
        upc=None,
        model_description="Media Server",
        model_number="v0.0.1",
        serial_number="0000001",
        presentation_url=None,
        url="/device.xml",
        icons=[],
        xml=ET.Element("server_device"),
    )
    EMBEDDED_DEVICES = []
    SERVICES = [ContentDirectoryService]

    def __init__(self, requester: UpnpRequester, base_uri: str, boot_id: int, config_id: int) -> None:
        """Initialize."""
        super().__init__(
            requester=requester,
            base_uri=base_uri,
            boot_id=boot_id,
            config_id=config_id,
        )



async def async_main(server: UpnpServer) -> None:
    """Main."""
    await server.async_start()

    while True:
        await asyncio.sleep(3600)


async def async_stop(server: UpnpServer) -> None:
    await server.async_stop()

    loop = asyncio.get_event_loop()
    loop.run_until_complete()


if __name__ == "__main__":
    boot_id = int(time())
    config_id = 1
    server = UpnpServer(MediaServerDevice, SOURCE, http_port=HTTP_PORT, boot_id=boot_id, config_id=config_id)

    try:
        asyncio.run(async_main(server))
    except KeyboardInterrupt:
        print(KeyboardInterrupt)

    asyncio.run(server.async_stop())



"""Unit tests for ssdp_listener."""

import asyncio
from datetime import datetime, timedelta
from typing import AsyncGenerator
from unittest.mock import ANY, AsyncMock, Mock, patch

import pytest

from async_upnp_client.advertisement import SsdpAdvertisementListener
from async_upnp_client.const import NotificationSubType, SsdpSource
from async_upnp_client.search import SsdpSearchListener
from async_upnp_client.ssdp_listener import (
    SsdpDevice,
    SsdpListener,
    is_valid_location,
    same_headers_differ,
)
from async_upnp_client.utils import CaseInsensitiveDict

from .common import (
    ADVERTISEMENT_HEADERS_DEFAULT,
    ADVERTISEMENT_REQUEST_LINE,
    SEARCH_HEADERS_DEFAULT,
    SEARCH_REQUEST_LINE,
)

UDN = ADVERTISEMENT_HEADERS_DEFAULT["_udn"]


@pytest.fixture(autouse=True)
async def mock_start_listeners() -> AsyncGenerator:
    """Create listeners but don't call async_start()."""
    # pylint: disable=protected-access

    async def async_start(self: SsdpListener) -> None:
        self._advertisement_listener = SsdpAdvertisementListener(
            on_alive=self._on_alive,
            on_update=self._on_update,
            on_byebye=self._on_byebye,
            source=self.source,
            target=self.target,
            loop=self.loop,
        )
        # await self._advertisement_listener.async_start()

        self._search_listener = SsdpSearchListener(
            callback=self._on_search,
            loop=self.loop,
            source=self.source,
            target=self.target,
            timeout=self.search_timeout,
        )
        # await self._search_listener.async_start()

    with patch.object(SsdpListener, "async_start", new=async_start) as mock:
        yield mock


async def see_advertisement(ssdp_listener: SsdpListener, request_line: str, headers: CaseInsensitiveDict) -> None:
    """See advertisement."""
    # pylint: disable=protected-access
    advertisement_listener = ssdp_listener._advertisement_listener
    assert advertisement_listener is not None
    advertisement_listener._on_data(request_line, headers)
    await asyncio.sleep(0)  # Allow callback to run, if called.


async def see_search(ssdp_listener: SsdpListener, request_line: str, headers: CaseInsensitiveDict) -> None:
    """See search."""
    # pylint: disable=protected-access
    search_listener = ssdp_listener._search_listener
    assert search_listener is not None
    search_listener._on_data(request_line, headers)
    await asyncio.sleep(0)  # Allow callback to run, if called.


@pytest.mark.parametrize(
    "location",
    [
        "http://127.0.0.1/desc.xml",
        "http://127.0.0.2/desc.xml",  # whole 127.0.0.0/8 is loopback, not just .1
        "http://localhost/desc.xml",
        "http://LocalHost/desc.xml",
        "http://[::1]/desc.xml",
        "http://169.254.169.254/latest/",
        # Evasions of the old substring filter that still target loopback/link-local:
        "http://[::ffff:127.0.0.1]/desc.xml",  # IPv4-mapped IPv6 loopback
        "http://[::ffff:7f00:1]/desc.xml",  # IPv4-mapped IPv6 loopback (hex)
        "http://user@127.0.0.1/desc.xml",  # credentials in authority
        "http://0.0.0.0/desc.xml",  # unspecified address
        "ftp://192.168.1.1/desc.xml",  # not http(s)
        # Obfuscated IPv4 encodings that resolve to loopback/link-local:
        "http://2130706433/desc.xml",  # decimal 127.0.0.1
        "http://0177.0.0.1/desc.xml",  # octal 127.0.0.1
        "http://0x7f.0.0.1/desc.xml",  # hex 127.0.0.1
        "http://127.1/desc.xml",  # short form 127.0.0.1
        "http://%31%32%37.0.0.1/desc.xml",  # percent-encoded 127.0.0.1
        "http://2852039166/desc.xml",  # decimal 169.254.169.254
        # Reserved/conventional loopback names and trailing-dot forms:
        "http://127.0.0.1./desc.xml",  # trailing root-label dot
        "http://localhost./desc.xml",
        "http://localhost.localdomain/desc.xml",
        "http://localhost.localdomain./desc.xml",
        "http://evil.localhost/desc.xml",  # RFC 6761 reserves *.localhost
        "http://foo.bar.localhost/desc.xml",
        # Malformed URLs must be rejected, not raise:
        "http://[::1/desc.xml",  # unbalanced IPv6 bracket
        "http://[bad/desc.xml",
    ],
)
def test_is_valid_location_rejects(location: str) -> None:
    """Locations resolving to loopback/link-local/unspecified must be rejected."""
    is_valid_location.cache_clear()
    assert is_valid_location(location) is False


@pytest.mark.parametrize(
    "location",
    [
        "http://192.168.1.1/desc.xml",  # RFC1918 — normal LAN device
        "http://10.0.0.1/desc.xml",
        "http://172.16.0.1/desc.xml",
        "http://[fc00::1]/desc.xml",  # IPv6 ULA — normal LAN device
        "http://[fd00::1]/desc.xml",
        "http://[fe80::1]/desc.xml",  # IPv6 link-local — normal LAN device
        "http://dlna_dmr:1234/device.xml",  # DNS name
        "http://my-device.local./desc.xml",  # trailing-dot FQDN, not loopback
        "https://192.168.1.1:49152/desc.xml",
    ],
)
def test_is_valid_location_accepts(location: str) -> None:
    """Normal LAN locations (RFC1918, ULA, DNS names) must be accepted."""
    is_valid_location.cache_clear()
    assert is_valid_location(location) is True


@pytest.mark.asyncio
async def test_see_advertisement_alive() -> None:
    """Test seeing a device through an ssdp:alive-advertisement."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through alive-advertisement.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_ALIVE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through alive-advertisement, not triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_not_awaited()
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_advertisement_byebye() -> None:
    """Test seeing a device through an ssdp:byebye-advertisement."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through byebye-advertisement, not triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:byebye"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_not_awaited()
    assert UDN not in listener.devices

    # See device for the first time through alive-advertisement, triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_ALIVE,
    )
    assert async_callback.await_args is not None
    device, dst, _ = async_callback.await_args.args
    assert device.combined_headers(dst)["NTS"] == NotificationSubType.SSDP_ALIVE
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through byebye-advertisement, triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:byebye"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_BYEBYE,
    )
    assert async_callback.await_args is not None
    device, dst, _ = async_callback.await_args.args
    assert device.combined_headers(dst)["NTS"] == NotificationSubType.SSDP_BYEBYE
    assert UDN not in listener.devices

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_advertisement_update() -> None:
    """Test seeing a device through a ssdp:update-advertisement."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through alive-advertisement, triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_ALIVE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through update-advertisement, triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:update"
    headers["BOOTID.UPNP.ORG"] = "2"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_UPDATE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search() -> None:
    """Test seeing a device through an search."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See same device again through search, not triggering a change.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_ALIVE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_sync() -> None:
    """Test seeing a device through an search."""
    # pylint: disable=protected-access
    callback = Mock()
    listener = SsdpListener(callback=callback)
    await listener.async_start()

    # See device for the first time through search.
    callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    callback.assert_called_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See same device again through search, not triggering a change.
    callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    callback.assert_called_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_ALIVE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_then_alive() -> None:
    """Test seeing a device through a search, then a ssdp:alive-advertisement."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through alive-advertisement, not triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_not_awaited()
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_then_update() -> None:
    """Test seeing a device through a search, then a ssdp:update-advertisement."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through update-advertisement, triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:update"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_UPDATE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_then_byebye() -> None:
    """Test seeing a device through a search, then a ssdp:byebye-advertisement."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through byebye-advertisement,
    # triggering byebye-callback and device removed.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:byebye"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_BYEBYE,
    )
    assert UDN not in listener.devices

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_then_byebye_then_alive() -> None:
    """Test seeing a device by search, then ssdp:byebye, then ssdp:alive."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through byebye-advertisement,
    # triggering byebye-callback and device removed.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:byebye"
    headers["LOCATION"] = ""
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_BYEBYE,
    )
    assert UDN not in listener.devices

    # See device for the second time through alive-advertisement, not triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(ADVERTISEMENT_HEADERS_DEFAULT)
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.ADVERTISEMENT_ALIVE,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    await listener.async_stop()


@pytest.mark.asyncio
async def test_purge_devices() -> None:
    """Test if a device is purged when it times out given the value of the CACHE-CONTROL header."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # "Wait" a bit... and purge devices.
    override_now = headers["_timestamp"] + timedelta(hours=1)
    listener._device_tracker.purge_devices(override_now)
    assert UDN not in listener.devices

    await listener.async_stop()


@pytest.mark.asyncio
async def test_purge_devices_2() -> None:
    """Test if a device is purged when it times out, part 2."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See anotherdevice through search.
    async_callback.reset_mock()
    udn2 = "uuid:device_2"
    new_timestamp = SEARCH_HEADERS_DEFAULT["_timestamp"] + timedelta(hours=1)
    device_2_headers = CaseInsensitiveDict(
        {
            **SEARCH_HEADERS_DEFAULT,
            "USN": udn2 + "::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:2",
            "ST": "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:2",
            "_udn": udn2,
            "_timestamp": new_timestamp,
        }
    )
    await see_search(listener, SEARCH_REQUEST_LINE, device_2_headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:2",
        SsdpSource.SEARCH_CHANGED,
    )
    assert UDN not in listener.devices
    assert udn2 in listener.devices
    assert listener.devices[udn2].location is not None

    await listener.async_stop()


def test_same_headers_differ_profile() -> None:
    """Test same_headers_differ."""
    current_headers = CaseInsensitiveDict(
        {
            "Cache-Control": "max-age=1900",
            "location": "http://192.168.1.1:80/RootDevice.xml",
            "Server": "UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0",
            "ST": "urn:schemas-upnp-org:device:WANDevice:1",
            "USN": "uuid:upnp-WANDevice-1_0-123456789abc::urn:schemas-upnp-org:device:WANDevice:1",
            "EXT": "",
            "_location_original": "http://192.168.1.1:80/RootDevice.xml",
            "_timestamp": datetime.now(),
            "_host": "192.168.1.1",
            "_port": "1900",
            "_udn": "uuid:upnp-WANDevice-1_0-123456789abc",
            "_source": SsdpSource.SEARCH,
        }
    )
    new_headers = CaseInsensitiveDict(
        {
            "Cache-Control": "max-age=1900",
            "location": "http://192.168.1.1:80/RootDevice.xml",
            "Server": "UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0 abc",
            "Date": "Sat, 11 Sep 2021 12:00:00 GMT",
            "ST": "urn:schemas-upnp-org:device:WANDevice:1",
            "USN": "uuid:upnp-WANDevice-1_0-123456789abc::urn:schemas-upnp-org:device:WANDevice:1",
            "EXT": "",
            "_location_original": "http://192.168.1.1:80/RootDevice.xml",
            "_timestamp": datetime.now(),
            "_host": "192.168.1.1",
            "_port": "1900",
            "_udn": "uuid:upnp-WANDevice-1_0-123456789abc",
            "_source": SsdpSource.SEARCH,
        }
    )
    for _ in range(0, 10000):
        assert not same_headers_differ(current_headers, new_headers)


@pytest.mark.asyncio
async def test_see_search_invalid_usn() -> None:
    """Test invalid USN is ignored."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()
    advertisement_listener = listener._advertisement_listener
    assert advertisement_listener is not None

    # See device for the first time through alive-advertisement.
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    headers["ST"] = "urn:Microsoft Windows Peer Name Resolution Protocol: V4:IPV6:LinkLocal"
    headers["USN"] = "[fe80::aaaa:bbbb:cccc:dddd]:3540"
    del headers["_udn"]
    advertisement_listener._on_data(SEARCH_REQUEST_LINE, headers)
    async_callback.assert_not_awaited()

    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_invalid_location() -> None:
    """Test headers with invalid location is ignored."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()
    advertisement_listener = listener._advertisement_listener
    assert advertisement_listener is not None

    # See device for the first time through alive-advertisement.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    headers["location"] = "192.168.1.1"
    advertisement_listener._on_data(SEARCH_REQUEST_LINE, headers)
    async_callback.assert_not_awaited()

    await listener.async_stop()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "location",
    [
        "http://127.0.0.1:1234/device.xml",
        "http://127.0.0.20:1234/device.xml",
        "http://[::1]:1234/device.xml",
        "http://169.254.12.1:1234/device.xml",
        "http://localhost:1234/device.xml",
    ],
)
async def test_see_search_localhost_location(location: str) -> None:
    """Test localhost location (127.0.0.1/[::1]) is ignored."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()
    advertisement_listener = listener._advertisement_listener
    assert advertisement_listener is not None

    # See device for the first time through alive-advertisement.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict(SEARCH_HEADERS_DEFAULT)
    headers["location"] = location
    advertisement_listener._on_data(SEARCH_REQUEST_LINE, headers)
    async_callback.assert_not_awaited()

    await listener.async_stop()


@pytest.mark.asyncio
async def test_combined_headers() -> None:
    """Test combined headers."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device for the first time through search.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict({**SEARCH_HEADERS_DEFAULT, "booTID.UPNP.ORG": "0", "Original": "2"})
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(
        ANY,
        "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        SsdpSource.SEARCH_CHANGED,
    )
    assert async_callback.await_args is not None
    device, dst, _ = async_callback.await_args.args
    assert UDN in listener.devices
    assert listener.devices[UDN].location is not None

    # See device for the second time through alive-advertisement, not triggering callback.
    async_callback.reset_mock()
    headers = CaseInsensitiveDict({**ADVERTISEMENT_HEADERS_DEFAULT, "BooTID.UPNP.ORG": "2"})
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)

    assert isinstance(device, SsdpDevice)
    combined = device.combined_headers(dst)
    assert isinstance(combined, CaseInsensitiveDict)
    result = {k.lower(): str(v) for k, v in combined.as_dict().items()}
    del result["_timestamp"]
    assert result == {
        "_host": "192.168.1.1",
        "_port": "1900",
        "_udn": "uuid:...",
        "bootid.upnp.org": "2",
        "cache-control": "max-age=1800",
        "date": "Fri, 1 Jan 2021 12:00:00 GMT",
        "location": "http://192.168.1.1:80/RootDevice.xml",
        "nt": "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        "nts": NotificationSubType.SSDP_ALIVE,
        "original": "2",
        "server": "Linux/2.0 UPnP/1.0 async_upnp_client/0.1",
        "st": "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        "usn": "uuid:...::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
    }
    assert combined["original"] == "2"
    assert combined["bootid.upnp.org"] == "2"
    assert "_source" not in combined

    headers = CaseInsensitiveDict(
        {
            **ADVERTISEMENT_HEADERS_DEFAULT,
            "BooTID.UPNP.ORG": "2",
            "st": "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:2",
        }
    )
    headers["NTS"] = "ssdp:alive"
    await see_advertisement(listener, ADVERTISEMENT_REQUEST_LINE, headers)
    combined = device.combined_headers(dst)
    assert combined["st"] == "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:2"
    await listener.async_stop()


@pytest.mark.asyncio
async def test_see_search_device_ipv4_and_ipv6() -> None:
    """Test seeing the same device via IPv4, then via IPv6."""
    # pylint: disable=protected-access
    async_callback = AsyncMock()
    listener = SsdpListener(async_callback=async_callback)
    await listener.async_start()

    # See device via IPv4, callback should be called.
    async_callback.reset_mock()
    location_ipv4 = "http://192.168.1.1:80/RootDevice.xml"
    headers = CaseInsensitiveDict(
        {
            **SEARCH_HEADERS_DEFAULT,
            "LOCATION": location_ipv4,
        }
    )
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(ANY, SEARCH_HEADERS_DEFAULT["ST"], SsdpSource.SEARCH_CHANGED)

    # See device via IPv6, callback should be called with SsdpSource.SEARCH_ALIVE,
    # not SEARCH_UPDATE.
    async_callback.reset_mock()
    location_ipv6 = "http://[fe80::1]:80/RootDevice.xml"
    headers = CaseInsensitiveDict(
        {
            **SEARCH_HEADERS_DEFAULT,
            "LOCATION": location_ipv6,
        }
    )
    await see_search(listener, SEARCH_REQUEST_LINE, headers)
    async_callback.assert_awaited_once_with(ANY, SEARCH_HEADERS_DEFAULT["ST"], SsdpSource.SEARCH_ALIVE)

    assert listener.devices[SEARCH_HEADERS_DEFAULT["_udn"]].locations



"""async_upnp_client.device_updater module."""

import logging

from async_upnp_client.advertisement import SsdpAdvertisementListener
from async_upnp_client.client import UpnpDevice
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import AddressTupleVXType
from async_upnp_client.utils import CaseInsensitiveDict

_LOGGER = logging.getLogger(__name__)


class DeviceUpdater:
    """
    Device updater.

    Listens for SSDP advertisements and updates device inline when needed.
    Inline meaning that it keeps the original UpnpDevice instance.
    So be sure to keep only references to the UpnpDevice,
    as a device might decide to remove a service after an update!
    """

    def __init__(
        self,
        device: UpnpDevice,
        factory: UpnpFactory,
        source: AddressTupleVXType | None = None,
    ) -> None:
        """Initialize."""
        self._device = device
        self._factory = factory
        self._listener = SsdpAdvertisementListener(
            async_on_alive=self._async_on_alive,
            async_on_byebye=self._async_on_byebye,
            async_on_update=self._async_on_update,
            source=source,
        )

    async def async_start(self) -> None:
        """Start listening for notifications."""
        _LOGGER.debug("Start listening for notifications.")
        await self._listener.async_start()

    async def async_stop(self) -> None:
        """Stop listening for notifications."""
        _LOGGER.debug("Stop listening for notifications.")
        await self._listener.async_stop()

    async def _async_on_alive(self, headers: CaseInsensitiveDict) -> None:
        """Handle on alive."""
        # Ensure for root devices only.
        if headers.get("nt") != "upnp:rootdevice":
            return

        # Ensure for our device.
        if headers.get("_udn") != self._device.udn:
            return

        _LOGGER.debug("Handling alive: %s", headers)
        await self._async_handle_alive_update(headers)

    async def _async_on_byebye(self, headers: CaseInsensitiveDict) -> None:
        """Handle on byebye."""
        _LOGGER.debug("Handling on_byebye: %s", headers)
        self._device.available = False

    async def _async_on_update(self, headers: CaseInsensitiveDict) -> None:
        """Handle on update."""
        # Ensure for root devices only.
        if headers.get("nt") != "upnp:rootdevice":
            return

        # Ensure for our device.
        if headers.get("_udn") != self._device.udn:
            return

        _LOGGER.debug("Handling update: %s", headers)
        await self._async_handle_alive_update(headers)

    async def _async_handle_alive_update(self, headers: CaseInsensitiveDict) -> None:
        """Handle on_alive or on_update."""
        do_reinit = False

        # Handle BOOTID.UPNP.ORG.
        boot_id = headers.get("BOOTID.UPNP.ORG")
        device_boot_id = self._device.ssdp_headers.get("BOOTID.UPNP.ORG")
        if boot_id and boot_id != device_boot_id:
            _LOGGER.debug("New boot_id: %s, old boot_id: %s", boot_id, device_boot_id)
            do_reinit = True

        # Handle CONFIGID.UPNP.ORG.
        config_id = headers.get("CONFIGID.UPNP.ORG")
        device_config_id = self._device.ssdp_headers.get("CONFIGID.UPNP.ORG")
        if config_id and config_id != device_config_id:
            _LOGGER.debug(
                "New config_id: %s, old config_id: %s",
                config_id,
                device_config_id,
            )
            do_reinit = True

        # Handle LOCATION.
        location = headers.get("LOCATION")
        if location and self._device.device_url != location:
            _LOGGER.debug("New location: %s, old location: %s", location, self._device.device_url)
            do_reinit = True

        if location and do_reinit:
            await self._reinit_device(location, headers)

        # We heard from it, so mark it available.
        self._device.available = True

    async def _reinit_device(self, location: str, ssdp_headers: CaseInsensitiveDict) -> None:
        """Reinitialize device."""
        # pylint: disable=protected-access
        _LOGGER.debug("Reinitializing device, location: %s", location)

        new_device = await self._factory.async_create_device(location)
        self._device.reinit(new_device)
        self._device.ssdp_headers = ssdp_headers



#!/usr/bin/env python3
"""Example of adding and deleting a port mapping."""

import asyncio
import ipaddress
import sys
from datetime import timedelta
from typing import cast

from async_upnp_client.aiohttp import AiohttpRequester
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.profiles.igd import IgdDevice
from async_upnp_client.utils import CaseInsensitiveDict, get_local_ip

SOURCE = ("0.0.0.0", 0)


async def discover_igd_devices() -> set[CaseInsensitiveDict]:
    """Discover IGD devices."""
    # Do the search, this blocks for timeout (4 seconds, default).
    discoveries = await IgdDevice.async_search(source=SOURCE)
    if not discoveries:
        print("Could not find device")
        sys.exit(1)

    return discoveries


async def build_igd_device(discovery: CaseInsensitiveDict) -> IgdDevice:
    """Find and construct device."""
    location = discovery["location"]
    requester = AiohttpRequester()
    factory = UpnpFactory(requester, non_strict=True)
    device = await factory.async_create_device(description_url=location)
    return IgdDevice(device, None)


async def async_add_port_mapping(igd_device: IgdDevice) -> None:
    """Add port mapping."""
    external_ip_address = await igd_device.async_get_external_ip_address()
    if not external_ip_address:
        print("Could not get external IP address")
        sys.exit(1)

    remote_host = ipaddress.ip_address(external_ip_address)
    remote_host_ipv4 = cast(ipaddress.IPv4Address, remote_host)
    local_ip = ipaddress.ip_address(get_local_ip())
    local_ip_ipv4 = cast(ipaddress.IPv4Address, local_ip)
    # Change `enabled` to False to disable port mapping.
    # NB: This does not delete the port mapping.
    enabled = True
    mapping_name = "Bombsquad"
    await igd_device.async_add_port_mapping(
        remote_host=remote_host_ipv4,
        external_port=43210,
        internal_client=local_ip_ipv4,
        internal_port=43210,
        protocol="UDP",
        enabled=enabled,
        description=mapping_name,
        lease_duration=timedelta(seconds=7200),
    )  # Time in secs


async def async_del_port_mapping(igd_device: IgdDevice) -> None:
    """Delete port mapping."""
    external_ip_address = await igd_device.async_get_external_ip_address()
    if not external_ip_address:
        print("Could not get external IP address")
        sys.exit(1)

    remote_host = ipaddress.ip_address(external_ip_address)
    remote_host_ipv4 = cast(ipaddress.IPv4Address, remote_host)
    await igd_device.async_delete_port_mapping(
        remote_host=remote_host_ipv4,
        external_port=43210,
        protocol="UDP",
    )


async def async_main() -> None:
    """Async main."""
    discoveries = await discover_igd_devices()
    print(f"Discoveries: {discoveries}")
    discovery = list(discoveries)[0]
    print(f"Using device at location: {discovery['location']}")
    igd_device = await build_igd_device(discovery)

    print("Creating port mapping")
    await async_add_port_mapping(igd_device)

    await asyncio.sleep(5)

    print("Deleting port mapping")
    await async_del_port_mapping(igd_device)


def main() -> None:
    """Main."""
    try:
        asyncio.run(async_main())
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



"""Profiles for upnp_client."""

import asyncio
import os.path
from collections import deque
from copy import deepcopy
from typing import Deque, Mapping, MutableMapping, cast

from async_upnp_client.client import UpnpRequester
from async_upnp_client.const import AddressTupleVXType, HttpRequest, HttpResponse
from async_upnp_client.event_handler import UpnpEventHandler, UpnpNotifyServer


def read_file(filename: str) -> str:
    """Read file."""
    path = os.path.join("tests", "fixtures", filename)
    with open(path, encoding="utf-8") as file:
        return file.read()


class UpnpTestRequester(UpnpRequester):
    """Test requester."""

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        response_map: Mapping[tuple[str, str], HttpResponse],
    ) -> None:
        """Class initializer."""
        self.response_map: MutableMapping[tuple[str, str], HttpResponse] = deepcopy(cast(MutableMapping, response_map))
        self.exceptions: Deque[Exception | None] = deque()

    async def async_http_request(
        self,
        http_request: HttpRequest,
    ) -> HttpResponse:
        """Do a HTTP request."""
        await asyncio.sleep(0.01)

        if self.exceptions:
            exception = self.exceptions.popleft()
            if exception is not None:
                raise exception

        key = (http_request.method, http_request.url)
        if key not in self.response_map:
            raise KeyError(f"Request not in response map: {key}")

        return self.response_map[key]


RESPONSE_MAP: Mapping[tuple[str, str], HttpResponse] = {
    # DLNA/DMR
    ("GET", "http://dlna_dmr:1234/device.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/device.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/device_embedded.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/device_embedded.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/device_incomplete.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/device_incomplete.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/device_with_empty_descriptor.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/device_with_empty_descriptor.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/RenderingControl_1.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/RenderingControl_1.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/ConnectionManager_1.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/ConnectionManager_1.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/AVTransport_1.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/AVTransport_1.xml"),
    ),
    ("GET", "http://dlna_dmr:1234/Empty_Descriptor.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dmr/Empty_Descriptor.xml"),
    ),
    ("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/ConnectionManager1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-cm1", "timeout": "Second-175"},
        "",
    ),
    ("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1"): HttpResponse(
        200,
        {"sid": "uuid:dummy", "timeout": "Second-300"},
        "",
    ),
    ("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/AVTransport1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-avt1", "timeout": "Second-150"},
        "",
    ),
    ("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/QPlay"): HttpResponse(
        200,
        {"sid": "uuid:dummy-qp1", "timeout": "Second-150"},
        "",
    ),
    ("UNSUBSCRIBE", "http://dlna_dmr:1234/upnp/event/ConnectionManager1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-cm1"},
        "",
    ),
    ("UNSUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1"): HttpResponse(
        200,
        {"sid": "uuid:dummy"},
        "",
    ),
    ("UNSUBSCRIBE", "http://dlna_dmr:1234/upnp/event/AVTransport1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-avt1"},
        "",
    ),
    ("UNSUBSCRIBE", "http://dlna_dmr:1234/upnp/event/QPlay"): HttpResponse(
        200,
        {"sid": "uuid:dummy-qp1"},
        "",
    ),
    # DLNA/DMS
    ("GET", "http://dlna_dms:1234/device.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dms/device.xml"),
    ),
    ("GET", "http://dlna_dms:1234/ConnectionManager_1.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dms/ConnectionManager_1.xml"),
    ),
    ("GET", "http://dlna_dms:1234/ContentDirectory_1.xml"): HttpResponse(
        200,
        {},
        read_file("dlna/dms/ContentDirectory_1.xml"),
    ),
    ("SUBSCRIBE", "http://dlna_dms:1234/upnp/event/ConnectionManager1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-cm1", "timeout": "Second-150"},
        "",
    ),
    ("SUBSCRIBE", "http://dlna_dms:1234/upnp/event/ContentDirectory1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-cd1", "timeout": "Second-150"},
        "",
    ),
    ("UNSUBSCRIBE", "http://dlna_dms:1234/upnp/event/ConnectionManager1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-cm1"},
        "",
    ),
    ("UNSUBSCRIBE", "http://dlna_dms:1234/upnp/event/ContentDirectory1"): HttpResponse(
        200,
        {"sid": "uuid:dummy-cd1"},
        "",
    ),
    # IGD
    ("GET", "http://igd:1234/device.xml"): HttpResponse(200, {}, read_file("igd/device.xml")),
    ("GET", "http://igd:1234/Layer3Forwarding.xml"): HttpResponse(
        200,
        {},
        read_file("igd/Layer3Forwarding.xml"),
    ),
    ("GET", "http://igd:1234/WANCommonInterfaceConfig.xml"): HttpResponse(
        200,
        {},
        read_file("igd/WANCommonInterfaceConfig.xml"),
    ),
    ("GET", "http://igd:1234/WANIPConnection.xml"): HttpResponse(
        200,
        {},
        read_file("igd/WANIPConnection.xml"),
    ),
}


class UpnpTestNotifyServer(UpnpNotifyServer):
    """Test notify server."""

    def __init__(
        self,
        requester: UpnpRequester,
        source: AddressTupleVXType,
        callback_url: str | None = None,
    ) -> None:
        """Initialize."""
        self._requester = requester
        self._source = source
        self._callback_url = callback_url
        self.event_handler = UpnpEventHandler(self, requester)

    @property
    def callback_url(self) -> str:
        """Return callback URL on which we are callable."""
        return self._callback_url or f"http://{self._source[0]}:{self._source[1]}/notify"

    async def async_start_server(self) -> None:
        """Start the server."""

    async def async_stop_server(self) -> None:
        """Stop the server."""
        await self.event_handler.async_unsubscribe_all()



"""Unit tests for the IGD profile."""

from datetime import datetime, timedelta

import pytest

from async_upnp_client.client import UpnpDevice
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import HttpResponse
from async_upnp_client.event_handler import UpnpEventHandler
from async_upnp_client.profiles.igd import (
    IgdDevice,
    IgdStateItem,
    TrafficCounterState,
)

from ..conftest import RESPONSE_MAP, UpnpTestNotifyServer, UpnpTestRequester, read_file


class FixedCounterIgdDevice(IgdDevice):
    """IGD device with deterministic counters for overflow tests."""

    def __init__(
        self,
        device: UpnpDevice,
        event_handler: UpnpEventHandler | None,
        initial_traffic_state: TrafficCounterState,
        current_traffic_state: TrafficCounterState,
    ) -> None:
        """Initialize test device with fixed traffic counters."""
        super().__init__(device, event_handler)
        self._last_traffic_state = initial_traffic_state
        self._current_traffic_state = current_traffic_state

    async def async_poll_traffic_data(
        self,
        items: set[IgdStateItem],
    ) -> TrafficCounterState:
        """Return a fixed traffic state for tests."""
        return self._current_traffic_state


@pytest.mark.asyncio
async def test_init_igd_profile() -> None:
    """Test if a IGD device can be initialized."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://igd:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = IgdDevice(device, event_handler=event_handler)
    assert profile


@pytest.mark.asyncio
async def test_get_total_bytes_received() -> None:
    """Test getting total bytes received."""
    responses = dict(RESPONSE_MAP)
    responses[("POST", "http://igd:1234/WANCommonInterfaceConfig")] = HttpResponse(
        200,
        {},
        read_file("igd/action_WANCIC_GetTotalBytesReceived.xml"),
    )
    requester = UpnpTestRequester(responses)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://igd:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = IgdDevice(device, event_handler=event_handler)
    total_bytes_received = await profile.async_get_total_bytes_received()
    assert total_bytes_received == 1337


@pytest.mark.asyncio
async def test_get_total_packets_received_empty_response() -> None:
    """Test getting total packets received with empty response, for broken (Draytek) device."""
    responses = dict(RESPONSE_MAP)
    responses[("POST", "http://igd:1234/WANCommonInterfaceConfig")] = HttpResponse(
        200,
        {},
        read_file("igd/action_WANCIC_GetTotalPacketsReceived.xml"),
    )
    requester = UpnpTestRequester(responses)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://igd:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = IgdDevice(device, event_handler=event_handler)
    total_bytes_received = await profile.async_get_total_packets_received()
    assert total_bytes_received is None


@pytest.mark.asyncio
async def test_get_status_info_invalid_uptime() -> None:
    """Test getting status info with an invalid uptime response."""
    responses = dict(RESPONSE_MAP)
    responses[("POST", "http://igd:1234/WANIPConnection")] = HttpResponse(
        200,
        {},
        read_file("igd/action_WANIPConnection_GetStatusInfoInvalidUptime.xml"),
    )
    requester = UpnpTestRequester(responses)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://igd:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = IgdDevice(device, event_handler=event_handler)
    status_info = await profile.async_get_status_info()
    assert status_info is None


@pytest.mark.asyncio
async def test_negative_bytes_received_counter() -> None:
    """
    Test getting a negative total bytes received counter.

    Some devices implement the counter as a signed integer (i4),
    which can result in negative values.
    """
    responses = dict(RESPONSE_MAP)
    responses[("POST", "http://igd:1234/WANCommonInterfaceConfig")] = HttpResponse(
        200,
        {},
        read_file("igd/action_WANCIC_GetTotalBytesReceived_i4.xml"),
    )
    requester = UpnpTestRequester(responses)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://igd:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = IgdDevice(device, event_handler=event_handler)
    total_bytes_received = await profile.async_get_total_bytes_received()
    assert total_bytes_received == 1615498126  # 2**31 + -531985522


@pytest.mark.asyncio
async def test_derived_rates_with_uint32_overflow() -> None:
    """Test derived rates with and without uint32 overflow handling."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://igd:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler

    previous_timestamp = datetime(2000, 1, 1, 0, 0, 0)
    initial_traffic_state = TrafficCounterState(
        timestamp=previous_timestamp,
        bytes_received=2**32 - 100,
        bytes_sent=2**32 - 200,
        packets_received=2**32 - 50,
        packets_sent=2**32 - 1,
        bytes_received_original=2**32 - 100,
        bytes_sent_original=2**32 - 200,
        packets_received_original=2**32 - 50,
        packets_sent_original=2**32 - 1,
    )
    current_traffic_state = TrafficCounterState(
        timestamp=previous_timestamp + timedelta(seconds=10),
        bytes_received=50,
        bytes_sent=25,
        packets_received=10,
        packets_sent=4,
        bytes_received_original=50,
        bytes_sent_original=25,
        packets_received_original=10,
        packets_sent_original=4,
    )
    profile = FixedCounterIgdDevice(
        device,
        event_handler,
        initial_traffic_state=initial_traffic_state,
        current_traffic_state=current_traffic_state,
    )

    state = await profile.async_get_traffic_and_status_data(
        items={
            IgdStateItem.KIBIBYTES_PER_SEC_RECEIVED,
            IgdStateItem.KIBIBYTES_PER_SEC_SENT,
            IgdStateItem.PACKETS_PER_SEC_RECEIVED,
            IgdStateItem.PACKETS_PER_SEC_SENT,
            IgdStateItem.KIBIBYTES_PER_SEC_RECEIVED_NO_ROLLOVER,
            IgdStateItem.KIBIBYTES_PER_SEC_SENT_NO_ROLLOVER,
            IgdStateItem.PACKETS_PER_SEC_RECEIVED_NO_ROLLOVER,
            IgdStateItem.PACKETS_PER_SEC_SENT_NO_ROLLOVER,
        },
        force_poll=True,
    )

    assert state.kibibytes_per_sec_received_no_rollover is None
    assert state.kibibytes_per_sec_sent_no_rollover is None
    assert state.packets_per_sec_received_no_rollover is None
    assert state.packets_per_sec_sent_no_rollover is None

    elapsed_seconds = (current_traffic_state.timestamp - initial_traffic_state.timestamp).total_seconds()

    assert state.kibibytes_per_sec_received == pytest.approx((150 / 1024) / elapsed_seconds)
    assert state.kibibytes_per_sec_sent == pytest.approx((225 / 1024) / elapsed_seconds)
    assert state.packets_per_sec_received == pytest.approx(60 / elapsed_seconds)
    assert state.packets_per_sec_sent == pytest.approx(5 / elapsed_seconds)



"""Test server functionality."""

import asyncio
import socket
import xml.etree.ElementTree as ET
from contextlib import asynccontextmanager, suppress
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    NamedTuple,
    cast,
)
from unittest.mock import Mock

import aiohttp
import pytest
import pytest_asyncio
from aiohttp.test_utils import TestClient
from aiohttp.web import Application, Request
from pytest_aiohttp.plugin import AiohttpClient

import async_upnp_client.aiohttp
import async_upnp_client.server
from async_upnp_client.client import UpnpStateVariable
from async_upnp_client.const import DeviceInfo, ServiceInfo
from async_upnp_client.server import (
    UpnpServer,
    UpnpServerDevice,
    UpnpServerService,
    callable_action,
    create_event_var,
    create_state_var,
)
from async_upnp_client.utils import CaseInsensitiveDict

from .conftest import read_file


class ServerServiceTest(UpnpServerService):
    """Test Service."""

    SERVICE_DEFINITION = ServiceInfo(
        service_id="urn:upnp-org:serviceId:TestServerService",
        service_type="urn:schemas-upnp-org:service:TestServerService:1",
        control_url="/upnp/control/TestServerService",
        event_sub_url="/upnp/event/TestServerService",
        scpd_url="/ContentDirectory.xml",
        xml=ET.Element("server_service"),
    )

    STATE_VARIABLE_DEFINITIONS = {
        "TestVariable_str": create_state_var("string"),
        "EventableTextVariable_ui4": create_event_var("ui4", default="0"),
        "A_ARG_TYPE_In_Var1_str": create_state_var("string"),
        "A_ARG_TYPE_In_Var2_ui4": create_state_var("ui4"),
    }

    @callable_action(
        name="SetValues",
        in_args={
            "In_Var1_str": "A_ARG_TYPE_In_Var1_str",
        },
        out_args={
            "TestVariable_str": "TestVariable_str",
            "EventableTextVariable_ui4": "EventableTextVariable_ui4",
        },
    )
    async def set_values(
        self,
        In_Var1_str: str,  # pylint: disable=invalid-name
    ) -> dict[str, UpnpStateVariable]:
        """Handle action."""
        self.state_variable("TestVariable_str").value = In_Var1_str
        return {
            "TestVariable_str": self.state_variable("TestVariable_str"),
            "EventableTextVariable_ui4": self.state_variable("EventableTextVariable_ui4"),
        }

    def set_eventable(self, value: int) -> None:
        """Eventable state-variable assignment."""
        event_var = self.state_variable("EventableTextVariable_ui4")
        event_var.value = value


class ServerDeviceTest(UpnpServerDevice):
    """Test device."""

    DEVICE_DEFINITION = DeviceInfo(
        device_type=":urn:schemas-upnp-org:device:TestServerDevice:1",
        friendly_name="Test Server",
        manufacturer="Test",
        manufacturer_url=None,
        model_name="TestServer",
        model_url=None,
        udn="uuid:adca2e25-cbe4-427a-a5c3-9b5931e7b79b",
        upc=None,
        model_description="Test Server",
        model_number="v0.0.1",
        serial_number="0000001",
        presentation_url=None,
        url="/device.xml",
        icons=[],
        xml=ET.Element("server_device"),
    )
    EMBEDDED_DEVICES = []
    SERVICES = [ServerServiceTest]


class AppRunnerMock:
    """Mock AppRunner."""

    # pylint: disable=too-few-public-methods

    def __init__(self, app: Any, *_args: Any, **_kwargs: Any) -> None:
        """Initialize."""
        self.app = app

    async def setup(self) -> None:
        """Configure AppRunner."""


class MockSocket:
    """Mock socket without 'bind'."""

    def __init__(self, sock: socket.socket) -> None:
        """Initialize."""
        self.sock = sock

    def bind(self, addr: Any) -> None:
        """Ignore bind."""

    def __getattr__(self, name: str) -> Any:
        """Passthrough."""
        return getattr(self.sock, name)


class Callback:
    """HTTP server to process callbacks."""

    def __init__(self) -> None:
        """Initialize."""
        self.callback: Callable[[aiohttp.web.Request], Awaitable[aiohttp.web.Response]] | None = None
        self.session: TestClient[Request, Application] | None = None
        self.app = aiohttp.web.Application()
        self.app.router.add_route("NOTIFY", "/{tail:.*}", self.handler)

    async def start(self, aiohttp_client: AiohttpClient) -> None:
        """Generate session."""
        self.session = await aiohttp_client(self.app)

    def set_callback(self, callback: Callable[[aiohttp.web.Request], Awaitable[aiohttp.web.Response]]) -> None:
        """Assign callback."""
        self.callback = callback

    async def handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        """Handle callback."""
        if self.callback:
            return await self.callback(request)  # pylint: disable=not-callable
        return aiohttp.web.Response(status=200)

    @asynccontextmanager
    async def ClientSession(self) -> AsyncIterator:  # pylint: disable=invalid-name
        """Test session."""
        if self.session:
            yield self.session


class UpnpServerTuple(NamedTuple):
    """Upnp server tuple."""

    http_client: TestClient
    ssdp_sockets: list[socket.socket]
    callback: Callback
    server: UpnpServer


@pytest_asyncio.fixture
async def upnp_server(monkeypatch: Any, aiohttp_client: AiohttpClient) -> AsyncGenerator[UpnpServerTuple, None]:
    """Fixture to initialize device."""
    # pylint: disable=too-few-public-methods

    ssdp_sockets: list[socket.socket] = []
    http_client = None

    def get_ssdp_socket_mock(*_args: Any, **_kwargs: Any) -> tuple[MockSocket, None, None]:
        sock1, sock2 = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM)
        ssdp_sockets.append(sock2)
        return MockSocket(sock1), None, None

    class TCPSiteMock:
        """Mock TCP connection."""

        def __init__(self, runner: aiohttp.web.AppRunner, *_args: Any, **_kwargs: Any) -> None:
            self.app = runner.app
            self.name = "TCPSiteMock"

        async def start(self) -> Any:
            """Create HTTP server."""
            nonlocal http_client
            http_client = await aiohttp_client(self.app)
            return http_client

    callback = Callback()
    monkeypatch.setattr(async_upnp_client.server, "AppRunner", AppRunnerMock)
    monkeypatch.setattr(async_upnp_client.server, "TCPSite", TCPSiteMock)
    monkeypatch.setattr(async_upnp_client.server, "get_ssdp_socket", get_ssdp_socket_mock)
    monkeypatch.setattr(async_upnp_client.aiohttp, "ClientSession", callback.ClientSession)
    server = UpnpServer(ServerDeviceTest, ("127.0.0.1", 0), http_port=80, boot_id=1, config_id=1)
    await server.async_start()

    assert http_client
    assert aiohttp_client
    await callback.start(aiohttp_client)
    yield UpnpServerTuple(http_client, ssdp_sockets, callback, server)
    # await server.async_stop()
    for sock in ssdp_sockets:
        sock.close()


@pytest.mark.asyncio
async def test_init(upnp_server: UpnpServerTuple) -> None:
    """Test device query."""
    # pylint: disable=redefined-outer-name
    http_client = upnp_server.http_client
    response = await http_client.get("/device.xml")
    assert response.status == 200
    data = await response.text()
    assert data == read_file("server/device.xml").strip()


@pytest.mark.asyncio
async def test_action(upnp_server: UpnpServerTuple) -> None:
    """Test action execution."""
    # pylint: disable=redefined-outer-name
    http_client = upnp_server.http_client
    response = await http_client.post(
        "/upnp/control/TestServerService",
        data=read_file("server/action_request.xml"),
        headers={
            "content-type": 'text/xml; charset="utf-8"',
            "user-agent": "Linux/1.0 UPnP/1.1 test/1.0",
            "soapaction": "urn:schemas-upnp-org:service:TestServerService:1#SetValues",
        },
    )
    assert response.status == 200
    data = await response.text()
    assert data == read_file("server/action_response.xml").strip()


@pytest.mark.asyncio
async def test_subscribe(upnp_server: UpnpServerTuple) -> None:
    """Test subscription to server event."""
    # pylint: disable=redefined-outer-name
    event = asyncio.Event()
    expect = 0

    async def on_callback(request: aiohttp.web.Request) -> aiohttp.web.Response:
        nonlocal expect
        data = await request.read()
        assert data == read_file(f"server/subscribe_response_{expect}.xml").strip().encode()
        expect += 1
        event.set()
        return aiohttp.web.Response(status=200)

    http_client = upnp_server.http_client
    callback = upnp_server.callback
    server = upnp_server.server
    server_device = server._device  # pylint: disable=protected-access
    assert server_device
    service = cast(
        ServerServiceTest,
        server_device.service("urn:schemas-upnp-org:service:TestServerService:1"),
    )
    callback.set_callback(on_callback)
    response = await http_client.request(
        "SUBSCRIBE",
        "/upnp/event/TestServerService",
        headers={"CALLBACK": "</foo/bar>", "NT": "upnp:event", "TIMEOUT": "Second-30"},
    )
    assert response.status == 200
    data = await response.text()
    assert not data
    sid = response.headers.get("SID")
    assert sid
    with suppress(asyncio.TimeoutError):
        await asyncio.wait_for(event.wait(), 2)
    assert event.is_set()

    event.clear()
    while not service.get_subscriber(sid):
        await asyncio.sleep(0)
    service.set_eventable(1)
    with suppress(asyncio.TimeoutError):
        await asyncio.wait_for(event.wait(), 2)
    assert event.is_set()


def test_send_search_response_ok(upnp_server: UpnpServerTuple) -> None:
    """Test sending search response without any failure."""
    # pylint: disable=redefined-outer-name, protected-access
    server = upnp_server.server
    search_responser = server._search_responder
    assert search_responser
    assert search_responser._response_transport
    response_transport = cast(Mock, search_responser._response_transport)
    assert response_transport
    response_transport.sendto = Mock(side_effect=None)

    headers = CaseInsensitiveDict(
        {
            "HOST": "192.168.1.100",
            "man": '"ssdp:discover"',
            "st": "upnp:rootdevice",
            "_remote_addr": ("192.168.1.101", 31234),
        }
    )
    search_responser._on_data("M-SEARCH * HTTP/1.1", headers)

    response_transport.sendto.assert_called()


def test_send_search_response_oserror(upnp_server: UpnpServerTuple) -> None:
    """Test sending search response and failing, but the error is handled."""
    # pylint: disable=redefined-outer-name, protected-access
    server = upnp_server.server
    search_responser = server._search_responder
    assert search_responser
    assert search_responser._response_transport
    response_transport = cast(Mock, search_responser._response_transport)
    assert response_transport
    response_transport.sendto = Mock(side_effect=None)

    headers = CaseInsensitiveDict(
        {
            "HOST": "192.168.1.100",
            "man": '"ssdp:discover"',
            "st": "upnp:rootdevice",
            "_remote_addr": ("192.168.1.101", 31234),
        }
    )
    search_responser._on_data("M-SEARCH * HTTP/1.1", headers)

    response_transport.sendto.assert_called()



"""async_upnp_client.event_handler module."""

import asyncio
import logging
import weakref
from abc import ABC
from datetime import timedelta
from http import HTTPStatus
from ipaddress import ip_address
from typing import Callable, Type
from urllib.parse import urlparse

import defusedxml.ElementTree as DET

from async_upnp_client.client import UpnpDevice, UpnpRequester, UpnpService
from async_upnp_client.const import NS, HttpRequest, IPvXAddress, ServiceId
from async_upnp_client.exceptions import (
    UpnpConnectionError,
    UpnpError,
    UpnpResponseError,
    UpnpSIDError,
)
from async_upnp_client.utils import get_local_ip

_LOGGER = logging.getLogger(__name__)


def default_on_pre_notify(request: HttpRequest) -> HttpRequest:
    """Pre-notify hook."""
    # pylint: disable=unused-argument
    fixed_body = (request.body or "").rstrip(" \t\r\n\0")
    return HttpRequest(request.method, request.url, request.headers, fixed_body)


class UpnpNotifyServer(ABC):
    """
    Base Notify Server, which binds to a UpnpEventHandler.

    A single UpnpNotifyServer/UpnpEventHandler can be shared with multiple UpnpDevices.
    """

    @property
    def callback_url(self) -> str:
        """Return callback URL on which we are callable."""
        raise NotImplementedError()

    async def async_start_server(self) -> None:
        """Start the server."""
        raise NotImplementedError()

    async def async_stop_server(self) -> None:
        """Stop the server."""
        raise NotImplementedError()


class UpnpEventHandler:
    """
    Handles UPnP eventing.

    An incoming NOTIFY request should be pass to handle_notify().
    subscribe/resubscribe/unsubscribe handle subscriptions.

    When using a reverse proxy in combination with a event handler, you should use
    the option to override the callback url.

    A single UpnpNotifyServer/UpnpEventHandler can be shared with multiple UpnpDevices.
    """

    def __init__(
        self,
        notify_server: UpnpNotifyServer,
        requester: UpnpRequester,
        on_pre_notify: Callable[[HttpRequest], HttpRequest] = default_on_pre_notify,
    ) -> None:
        """
        Initialize.

        notify_server is the notify server which is actually listening on a socket.
        """
        self._notify_server = notify_server
        self._requester = requester
        self.on_pre_notify = on_pre_notify

        self._subscriptions: weakref.WeakValueDictionary[ServiceId, UpnpService] = weakref.WeakValueDictionary()
        self._backlog: dict[ServiceId, HttpRequest] = {}

    @property
    def callback_url(self) -> str:
        """Return callback URL on which we are callable."""
        return self._notify_server.callback_url

    def sid_for_service(self, service: UpnpService) -> ServiceId | None:
        """Get the service connected to SID."""
        for sid, subscribed_service in self._subscriptions.items():
            if subscribed_service == service:
                return sid

        return None

    def service_for_sid(self, sid: ServiceId) -> UpnpService | None:
        """Get a UpnpService for SID."""
        return self._subscriptions.get(sid)

    def _sid_and_service(self, service_or_sid: UpnpService | ServiceId) -> tuple[ServiceId, UpnpService]:
        """
        Resolve a SID or service to both SID and service.

        :raise KeyError: Cannot determine SID from UpnpService, or vice versa.
        """
        sid: ServiceId | None
        service: UpnpService | None

        if isinstance(service_or_sid, UpnpService):
            service = service_or_sid
            sid = self.sid_for_service(service)
            if not sid:
                raise KeyError(f"Unknown UpnpService {service}")
        else:
            sid = service_or_sid
            service = self.service_for_sid(sid)
            if not service:
                raise KeyError(f"Unknown SID {sid}")

        return sid, service

    async def handle_notify(self, http_request: HttpRequest) -> HTTPStatus:
        """Handle a NOTIFY request."""
        http_request = self.on_pre_notify(http_request)

        # ensure valid request
        if "NT" not in http_request.headers or "NTS" not in http_request.headers:
            return HTTPStatus.BAD_REQUEST

        if (
            http_request.headers["NT"] != "upnp:event"
            or http_request.headers["NTS"] != "upnp:propchange"
            or "SID" not in http_request.headers
        ):
            return HTTPStatus.PRECONDITION_FAILED

        sid: ServiceId = http_request.headers["SID"]
        service = self.service_for_sid(sid)

        # SID not known yet? store it in the backlog
        # Some devices don't behave nicely and send events before the SUBSCRIBE call is done.
        if not service:
            _LOGGER.debug("Storing NOTIFY in backlog for SID: %s", sid)
            self._backlog[sid] = http_request

            return HTTPStatus.OK

        # decode event and send updates to service
        changes = {}
        el_root = DET.fromstring(http_request.body)
        for el_property in el_root.findall("./event:property", NS):
            for el_state_var in el_property:
                name = el_state_var.tag
                value = el_state_var.text or ""
                changes[name] = value

        # send changes to service
        service.notify_changed_state_variables(changes)

        return HTTPStatus.OK

    async def async_subscribe(
        self,
        service: UpnpService,
        timeout: timedelta = timedelta(seconds=1800),
    ) -> tuple[ServiceId, timedelta]:
        """
        Subscription to a UpnpService.

        Be sure to re-subscribe before the subscription timeout passes.

        :param service: UpnpService to subscribe to self
        :param timeout: Timeout of subscription
        :return: SID (subscription ID), renewal timeout (may be different to
            supplied timeout)
        :raise UpnpResponseError: Error in response to subscription request
        :raise UpnpSIDError: No SID received for subscription
        :raise UpnpConnectionError: Device might be offline.
        :raise UpnpCommunicationError (or subclass): Error while performing
            subscription request.
        """
        _LOGGER.debug("Subscribing to: %s, callback URL: %s", service, self.callback_url)

        # do SUBSCRIBE request
        headers = {
            "NT": "upnp:event",
            "TIMEOUT": "Second-" + str(timeout.seconds),
            "HOST": urlparse(service.event_sub_url).netloc,
            "CALLBACK": f"<{self.callback_url}>",
        }
        backlog_request = HttpRequest("SUBSCRIBE", service.event_sub_url, headers, None)
        response = await self._requester.async_http_request(backlog_request)

        # check results
        if response.status_code != 200:
            _LOGGER.debug("Did not receive 200, but %s", response.status_code)
            raise UpnpResponseError(status=response.status_code, headers=response.headers)

        if "sid" not in response.headers:
            _LOGGER.debug("No SID received, aborting subscribe")
            raise UpnpSIDError

        # Device can give a different TIMEOUT header than what we have provided.
        if (
            "timeout" in response.headers
            and response.headers["timeout"] != "Second-infinite"
            and "Second-" in response.headers["timeout"]
        ):
            response_timeout = response.headers["timeout"]
            timeout_seconds = int(response_timeout[7:])  # len("Second-") == 7
            timeout = timedelta(seconds=timeout_seconds)

        sid: ServiceId = response.headers["sid"]
        self._subscriptions[sid] = service
        _LOGGER.debug("Subscribed, service: %s, SID: %s, timeout: %s", service, sid, timeout)

        # replay any backlog we have for this service
        if sid in self._backlog:
            _LOGGER.debug("Re-playing backlogged NOTIFY for SID: %s", sid)
            backlog_request = self._backlog[sid]
            await self.handle_notify(backlog_request)
            del self._backlog[sid]

        return sid, timeout

    async def _async_do_resubscribe(
        self,
        service: UpnpService,
        sid: ServiceId,
        timeout: timedelta = timedelta(seconds=1800),
    ) -> tuple[ServiceId, timedelta]:
        """Perform only a resubscribe, caller can retry subscribe if this fails."""
        # do SUBSCRIBE request
        headers = {
            "HOST": urlparse(service.event_sub_url).netloc,
            "SID": sid,
            "TIMEOUT": "Second-" + str(timeout.total_seconds()),
        }
        request = HttpRequest("SUBSCRIBE", service.event_sub_url, headers, None)
        response = await self._requester.async_http_request(request)

        # check results
        if response.status_code != 200:
            _LOGGER.debug("Did not receive 200, but %s", response.status_code)
            raise UpnpResponseError(status=response.status_code, headers=response.headers)

        # Devices should return the SID when re-subscribe,
        # but in case it doesn't, use the new SID.
        if "sid" in response.headers and response.headers["sid"]:
            new_sid: ServiceId = response.headers["sid"]
            if new_sid != sid:
                del self._subscriptions[sid]
                sid = new_sid

        # Device can give a different TIMEOUT header than what we have provided.
        if (
            "timeout" in response.headers
            and response.headers["timeout"] != "Second-infinite"
            and "Second-" in response.headers["timeout"]
        ):
            response_timeout = response.headers["timeout"]
            timeout_seconds = int(response_timeout[7:])  # len("Second-") == 7
            timeout = timedelta(seconds=timeout_seconds)

        self._subscriptions[sid] = service
        _LOGGER.debug("Resubscribed, service: %s, SID: %s, timeout: %s", service, sid, timeout)

        return sid, timeout

    async def async_resubscribe(
        self,
        service_or_sid: UpnpService | ServiceId,
        timeout: timedelta = timedelta(seconds=1800),
    ) -> tuple[ServiceId, timedelta]:
        """
        Renew subscription to a UpnpService.

        :param service_or_sid: UpnpService or existing SID to resubscribe
        :param timeout: Timeout of subscription
        :return: SID (subscription ID), renewal timeout (may be different to
            supplied timeout)
        :raise KeyError: Supplied service_or_sid is not known.
        :raise UpnpResponseError: Error in response to subscription request
        :raise UpnpSIDError: No SID received for subscription
        :raise UpnpConnectionError: Device might be offline.
        :raise UpnpCommunicationError (or subclass): Error while performing
            subscription request.
        """
        _LOGGER.debug("Resubscribing to: %s", service_or_sid)

        # Try a regular resubscribe. If that fails, delete old subscription and
        # do a full subscribe again.

        sid, service = self._sid_and_service(service_or_sid)
        try:
            return await self._async_do_resubscribe(service, sid, timeout)
        except UpnpConnectionError as err:
            _LOGGER.debug(
                "Resubscribe for %s failed: %s. Device offline, not retrying.",
                service_or_sid,
                err,
            )
            del self._subscriptions[sid]
            raise
        except UpnpError as err:
            _LOGGER.debug(
                "Resubscribe for %s failed: %s. Trying full subscribe.",
                service_or_sid,
                err,
            )
        del self._subscriptions[sid]
        return await self.async_subscribe(service, timeout)

    async def async_resubscribe_all(self) -> None:
        """Renew all current subscription."""
        await asyncio.gather(*(self.async_resubscribe(sid) for sid in self._subscriptions))

    async def async_unsubscribe(
        self,
        service_or_sid: UpnpService | ServiceId,
    ) -> ServiceId:
        """Unsubscribe from a UpnpService."""
        sid, service = self._sid_and_service(service_or_sid)

        _LOGGER.debug(
            "Unsubscribing from SID: %s, service: %s device: %s",
            sid,
            service,
            service.device,
        )

        # Remove registration before potential device errors
        del self._subscriptions[sid]

        # do UNSUBSCRIBE request
        headers = {
            "HOST": urlparse(service.event_sub_url).netloc,
            "SID": sid,
        }
        request = HttpRequest("UNSUBSCRIBE", service.event_sub_url, headers, None)
        response = await self._requester.async_http_request(request)

        # check results
        if response.status_code != 200:
            _LOGGER.debug("Did not receive 200, but %s", response.status_code)
            raise UpnpResponseError(status=response.status_code, headers=response.headers)

        return sid

    async def async_unsubscribe_all(self) -> None:
        """Unsubscribe all subscriptions."""
        sids = list(self._subscriptions)
        await asyncio.gather(
            *(self.async_unsubscribe(sid) for sid in sids),
            return_exceptions=True,
        )

    async def async_stop(self) -> None:
        """Stop event the UpnpNotifyServer."""
        # This calls async_unsubscribe_all() via the notify server.
        await self._notify_server.async_stop_server()


class UpnpEventHandlerRegister:
    """Event handler register to handle multiple interfaces."""

    def __init__(self, requester: UpnpRequester, notify_server_type: Type) -> None:
        """Initialize."""
        self.requester = requester
        self.notify_server_type = notify_server_type
        self._event_handlers: dict[IPvXAddress, tuple[UpnpEventHandler, set[UpnpDevice]]] = {}

    def _get_event_handler_for_device(self, device: UpnpDevice) -> UpnpEventHandler | None:
        """Get the event handler for the device, if known."""
        local_ip_str = get_local_ip(device.device_url)
        local_ip = ip_address(local_ip_str)
        if local_ip not in self._event_handlers:
            return None

        event_handler, devices = self._event_handlers[local_ip]
        if device in devices:
            return event_handler

        return None

    def has_event_handler_for_device(self, device: UpnpDevice) -> bool:
        """Check if an event handler for a device is already available."""
        return self._get_event_handler_for_device(device) is not None

    async def async_add_device(self, device: UpnpDevice) -> UpnpEventHandler:
        """Add a new device, creates or gets the event handler for this device."""
        local_ip_str = get_local_ip(device.device_url)
        local_ip = ip_address(local_ip_str)
        if local_ip not in self._event_handlers:
            event_handler = await self._create_event_handler_for_device(device)
            self._event_handlers[local_ip] = (event_handler, set([device]))
            return event_handler

        event_handler, devices = self._event_handlers[local_ip]
        devices.add(device)

        return event_handler

    async def _create_event_handler_for_device(self, device: UpnpDevice) -> UpnpEventHandler:
        """Create a new event handler for a device."""
        local_ip_str = get_local_ip(device.device_url)
        source_addr = (local_ip_str, 0)
        notify_server: UpnpNotifyServer = self.notify_server_type(requester=self.requester, source=source_addr)
        await notify_server.async_start_server()
        return UpnpEventHandler(notify_server, self.requester)

    async def async_remove_device(self, device: UpnpDevice) -> UpnpEventHandler | None:
        """Remove an existing device, destroys the event handler and returns it, if needed."""
        local_ip_str = get_local_ip(device.device_url)
        local_ip = ip_address(local_ip_str)
        assert local_ip in self._event_handlers

        event_handler, devices = self._event_handlers[local_ip]
        assert device in devices
        devices.remove(device)

        if not devices:
            await event_handler.async_stop()
            del self._event_handlers[local_ip]
            return event_handler

        return None
